/*$Id: getlistno.c,v 1.5 1999/01/21 03:16:20 lindberg Exp $*/
/*$Name: ezmlm-idx-0324 $*/
#include "stralloc.h"
#include "fmt.h"
#include "strerr.h"
#include "slurp.h"
#include "errtxt.h"
#include "subscribe.h"
#include <mysql/mysql.h>

static stralloc line = {0};
static stralloc quoted = {0};
static char strnum[FMT_ULONG];	/* message number as sz */

char *getlistno(dir,msgnum,listname)
/* gets dir/sql. If not found returns "". If found, populates stralloc listno */
/* with the lowest listno for the sublist named in dir/sql. If listname is set*/
/* it is used instead of the name from dir/sql.If there is an error, a pointer*/
/* to the error message is returned. Returns NULL on success. If no listno is */
/* found, listno is set to 0. This avoids further futile lookups. Still       */
/* returns "" in this case. */
char *dir;
unsigned long msgnum;	/* message number. Used to check for active sublist */
char * listname;	/* list name. This is needed so that ezmlm-receipt  */
			/* can get info also for sublists that mail receipts*/
			/* to it. For normal use this is NULL */
{
  MYSQL_RES *result;
  MYSQL_ROW row;
  unsigned long *lengths;
  char *table = "ezmlm";
  char *sublistname = (char *) 0;
  char *ret;
  unsigned int len;

  if (!listname)		/* if we spec listname we want a new one */
    if (listno.len) return (char *) 0;	/* otherwise, use the one we've got */
	/* We don't check dir against previous call. We assume that all SQL */
	/* using dbs of a list (e.g. subscribers, "allow", "digest" have the*/
	/* same "name"). If the dir/sql files have differnt names, this will*/
	/* get screwed up. Document! */
  if ((ret = opensql(dir,&table,&sublistname))) return ret;
  if (listname) sublistname = listname;
  if (!sublistname) {
    if (!stralloc_copys(&listno,"0")) return ERR_NOMEM;
    return "";
  }

  strnum[fmt_ulong(strnum,msgnum)] = '\0';		/* message nr ->string*/

	/* select listno from list_name where name='sublist' AND notuse = 0 */
	/* AND (msgnum BETWEEN msgnum_lo AND msgnum_hi) ORDER BY listno     */
	/* LIMIT 1. So, lowest listno matching 'sublist' and active for this*/
	/* message number. There may be more than one entry for a list, but */
	/* we always log to the lowest active listno */
  if (!stralloc_copys(&line,"SELECT listno FROM ")) return ERR_NOMEM;
  if (!stralloc_cats(&line,table)) return ERR_NOMEM;
  if (!stralloc_cats(&line,"_name WHERE name='")) return ERR_NOMEM;
  len = str_len(sublistname);
  if (!stralloc_ready(&quoted,2 * len + 1)) return ERR_NOMEM;
  quoted.len = mysql_escape_string(quoted.s,sublistname,len);
  if (!stralloc_cat(&line,&quoted)) return ERR_NOMEM;
  if (!stralloc_cats(&line,"' AND notuse = 0 AND (")) return ERR_NOMEM;
  if (!stralloc_cats(&line,strnum)) return ERR_NOMEM;
  if (!stralloc_cats(&line," BETWEEN msgnum_lo AND msgnum_hi)"
	" ORDER BY listno LIMIT 1"))
		return ERR_NOMEM;

  if (mysql_real_query((MYSQL *) psql,line.s,line.len) != 0)	/* lno query */
	return mysql_error((MYSQL *) psql);
  if (!(result = mysql_use_result((MYSQL *) psql)))		/* use result */
	return mysql_error((MYSQL *) psql);
  if (!(row = mysql_fetch_row(result))) {
    if (!mysql_eof(result))	/* eof => no entry. Otherwise error! */
	return mysql_error((MYSQL *) psql);
    else {			/* set to "0" if none found. */
	if (!stralloc_copys(&listno,"0")) return ERR_NOMEM;
        return "";
    }
  }
  if (! (lengths = mysql_fetch_lengths(result)))
	return mysql_error((MYSQL *) psql);
  if (!stralloc_copyb(&listno,row[0],lengths[0])) return ERR_NOMEM;
  mysql_free_result(result);				/* free result */
  return (char *) 0;
}

