/*$Id: tagmsg.c,v 1.9 1999/02/05 04:01:09 lindberg Exp $*/
/*$Name: ezmlm-idx-0324 $*/
#include "stralloc.h"
#include "scan.h"
#include "fmt.h"
#include "strerr.h"
#include "cookie.h"
#include "slurp.h"
#include "errtxt.h"
#include "subscribe.h"
#include <mysql/mysql.h>
#include <mysql/mysqld_error.h>

static stralloc line = {0};
static stralloc key = {0};
static char hash[COOKIE];
static char strnum[FMT_ULONG];	/* message number as sz */

static void die_nomem(fatal)
char *fatal;
{
  strerr_die2x(100,fatal,ERR_NOMEM);
}

void tagmsg(dir,msgnum,sender,action,flagsublist,hashout,fatal)
/* For a main list, this routine creates a cookie from num,sender and the */
/* list key and returns that cookie in hashout. The use of sender/num and */
/* first char of action is used to make cookie differ between messages,   */
/* the key is the secret list key. The cookie will be inserted into       */
/* table_cookie where table and other data is taken from dir/sql. If the  */
/* list is a sublist, we check if this list has already processed this    */
/* message. If so, it's a permanent error. This prevents replay. Next for */
/* both main and sublist, we log arrival of the message (done=0). */

char *dir;			/* db base dir */
unsigned long msgnum;		/* number of this message */
char *sender;			/* sender. NULL ok, but less entropy */
char *action;			/* to make it certain the cookie differs from*/
				/* one used for a digest */
int flagsublist;		/* = 1 <=> this is a sublist. 0 otherswise */
stralloc *hashout;		/* calculated hash goes here */
char *fatal;
{
  MYSQL_RES *result;
  MYSQL_ROW row;
  char *table = "ezmlm";
  char *sublistname = (char *) 0;
  char *ret;

  if ((ret = opensql(dir,&table,&sublistname))) {
    if (*ret) strerr_die2x(111,fatal,ret);
    return;
  }		/* do nothing if sql not in use */

  strnum[fmt_ulong(strnum,msgnum)] = '\0';	/* message nr ->string*/
  if (!sender) sender = strnum + 1;		/* always defined */

  if (sublistname)	/* get here only if opensql returned success. Thus, if*/
			/* getlistno returns "", it means that the list is not*/
			/* active, rather than that it's opensql returning """*/
    if ((ret = getlistno(dir,msgnum,(char *) 0)))
      if (*ret) strerr_die2x(111,fatal,ret);
      else strerr_die2x(100,fatal,ERR_NOT_ACTIVE);

  if (!flagsublist) {	/* main list. Create cookie */
    switch(slurp("key",&key,32)) {
      case -1:
	strerr_die3sys(111,fatal,ERR_READ,"key: ");
      case 0:
	strerr_die3x(100,fatal,"key",ERR_NOEXIST);
    }
    cookie(hash,key.s,key.len,strnum,sender,action);
    if (!stralloc_copyb(hashout,hash,COOKIE)) die_nomem(fatal);

	/* INSERT INTO table_cookie (msgnum,cookie) VALUES (num,cookie) */
	/* (we may have tried message before, but failed to complete, so */
	/* ER_DUP_ENTRY is ok) */
    if (!stralloc_copys(&line,"INSERT INTO ")) die_nomem(fatal);
    if (!stralloc_cats(&line,table)) die_nomem(fatal);
    if (!stralloc_cats(&line,"_cookie (msgnum,cookie) VALUES ("))
		die_nomem(fatal);
    if (!stralloc_cats(&line,strnum)) die_nomem(fatal);
    if (!stralloc_cats(&line,",'")) die_nomem(fatal);
    if (!stralloc_catb(&line,hash,COOKIE)) die_nomem(fatal);
    if (!stralloc_cats(&line,"')")) die_nomem(fatal);
    if (mysql_real_query((MYSQL *) psql,line.s,line.len) != 0)
      if (mysql_errno((MYSQL *) psql) != ER_DUP_ENTRY)	/* ignore dups */
        strerr_die2x(111,fatal,mysql_error((MYSQL *) psql)); /* cookie query */

  } else if (sublistname) {	/* If it's a sublist and we've got a way of  */
		/* identifying it, check that the message wasn't done already*/
		/* SELECT listno FROM table_mlog WHERE listno=ln AND msgnum  */
		/* = num and done = 1 [it may be 1/2 - done, -1 - bounced. If*/
		/* the latter, it's ok to try again.] Normally, there is no  */
		/* entry at all, since the list hasn't seen the msg before.  */
    if (!stralloc_copys(&line,"SELECT listno FROM ")) die_nomem(fatal);
    if (!stralloc_cats(&line,table)) die_nomem(fatal);
    if (!stralloc_cats(&line,"_mlog WHERE listno=")) die_nomem(fatal);
    if (!stralloc_cat(&line,&listno)) die_nomem(fatal);
    if (!stralloc_cats(&line," AND msgnum=")) die_nomem(fatal);
    if (!stralloc_cats(&line,strnum)) die_nomem(fatal);
    if (!stralloc_cats(&line," AND done > 0")) die_nomem(fatal);
    if (mysql_real_query((MYSQL *) psql,line.s,line.len) != 0)	/* query */
	strerr_die2x(111,fatal,mysql_error((MYSQL *) psql));
    if (!(result = mysql_use_result((MYSQL *) psql)))		/* use result */
	strerr_die2x(111,fatal,mysql_error((MYSQL *) psql));
    if ((row = mysql_fetch_row(result)))
	strerr_die2x(100,fatal,ERR_DONE);
    else							/* no result */
      if (!mysql_eof(result))
	strerr_die2x(111,fatal,mysql_error((MYSQL *) psql));
    mysql_free_result(result);					/* free res */
  }

  if (! (ret = logmsg(dir,msgnum,0))) return;			/* log done=0*/
  if (*ret) strerr_die2x(111,fatal,ret);
  return;
}
