#!/usr/bin/perl -T
# CGI program for retrieving sublist status info for a SQL-enabled ezmlm list
#
#	(c) 1999,	Frederik P. Lindberg
#			You may use under GPL
#
#			for more info, see http://www.ezmlm.org/
#
#
# Returns a status table for the last 10 messages if called without args.
# With appropriate args (supplied in the link from each status table entry)
# the program will explain in more detail.
#
# NOTE: You must at least change %DSOURCE below to make it work!
#
# TODO: -The constants determining the colors and the explanation should
#	automatically be adjusted [within a range] by past experience for
#	a list.
#	-Median or 90th percentile for deliveries. Stats for entire list
#	over range. Quick check returning which messages have problems.
#	Quick check on abnormal delays, i.e. messages within last 10 that
# 	are outside of 95% of previous messages.
#
# Logging is always to the lowest listno for
# a set of active list_name entries for this message. Without knowing
# this some of the code is very obscure. We just put in a pointer to
# the "looging" listno as the name of entries for the same name, but
# higher listno. We then suppress/handle differently these pointer
# entries at display. It's complicated by the fact that the range
# displayed may cover different configurations.
#
########################## CONFIGURE ######################################
### Link to this program
$URL='/cgi-bin/status.pl';

### Number of messages to display status info for
  $NOMSG = 10;

### Path for the colored balls that are used in the table. There should be
### $IMAGES/[red|green|yellow|orange|clear|white].gif.
  $IMAGES = '/images';

### DBI database driver
  $DRIVER = 'mysql';

### Datasource entries. We use PATH_INFO as the key to the hash, and if
### not all the info is there, we use the info from %DSOURCE{default} as
### defaults. You MUST supply a 'default' entry with all the info. If you
### use a non-default port, add ":port" to the host name.
### list => tableroot|db|host|user|pw
%DSOURCE = ( default => 'list|ezmlm|localhost|lindberg|gugugaga',
	list => 'list',
	digest => 'list_digest');

### Refresh interval for main display. Lots of viewers and short interval
### May load SQL server to slow list! Default => 300 seconds
  $REFRESH = 300;

###### Colors
### Done with the keyword to not clutter HTML unnecessarily when using default

### Page body background
  $BBGCOLOR="BGCOLOR=\"#F0F0A0\"";

### Text/Links/visited links
#  $TEXT="";
  $TEXT="TEXT=\"#000000\"";
  $VLINK="VLINK=\"#000000\"";
  $LINK="LINK=\"#000000\"";

### Table background
#  $TBGCOLOR="BGCOLOR=\"#FF8080\"";
  $TBGCOLOR="";

### Header background
  $HBGCOLOR="BGCOLOR=\"C0C0FF\"";

### Data background color
  $DBGCOLOR="BGCOLOR=\"#FFFFFF\"";

### Seconds to look back for messages. Since digests aren't consecutive,
### we use ORDER BY ... LIMIT. To not do too much ORDER BY work we limit
### how far back we go. Assume digest at least q 3 days.
  $DELTAT = $NOMSG * 3600 * 24 * 3;
 
############################################################################
### Check some config items
if (!$DSOURCE{'default'}) {		# Told you, but don't trust you
  print_error("default datasource is not defined");
} elsif ((split '\|', $DSOURCE{'default'}) != 5) {
  print_error("default datasource has wrong number of arguments");
}  

################## Prototypes ####################
sub content_type($;$);
sub clean_name($);
sub expl($$$$$);
sub status();
sub get_timenow($);
sub get_subno($);
sub stats();
sub list_split();
sub make_sent($;$);
sub make_name($$$);
sub make_timestring($);
sub db_connect();
sub msg_info($);
sub list_info($);
sub print_body();
sub print_error($);
sub print_head($;$);

my $msgmax,$magmin,$nomsg;
######### determine connect info explanded via DSOURCE{} ###################

  $pi = $ENV{'PATH_INFO'};
  $pi = $ARGV[1] if ($ARGV[1]);	# for testing
  $pi =~ s/^\///;		# delete first '/'
  $pi =~ s/\"//g;		# delete '"' as they might screw things up
  $ds = $DSOURCE{"$pi"};	# use 'default' entry if args don't match
  if (!$ds) { $pi = 'default'; }	# for headers

  ($troot,$db,$host,$user,$pw) = split '\|', $DSOURCE{"default"};
  my ($t,$d,$h,$u,$p) = split '\|',$ds;
  if ($t) { $troot = $t; }	# override defaults
  if ($d) { $db = $d; }
  if ($h) { $host = $h; }
  if ($u) { $user = $u; }
  if ($p) { $pw = $p; }

######### Determine in which function the program is called #################
  $qry = $ENV{'QUERY_STRING'};
  $pi = $ARGV[1] if ($ARGV[1]);	# for testing
  if ($qry =~ /^(\d+)_(\d+)_(-?\d+)_(-?\d+)_([-\w\._]+@?[-\w\._]+)$/) {
  		# digits: message nr
  		# digits: time since sent
  		# [-]digits: time since logged
  		# [-]digits: logged code
  		# (letters|digits|.|_)'@'(letters|digits|.|_): listname
    expl($1,$2,$3,$4,$5);
  } else {
    use DBI;	# others require database access
    if ($qry =~ /^(\d+)$/) {	# single numeric arg = message
      msg_info($1);
    } elsif ($qry =~ /^stats$/) {	# stats on last NOMSG messages
      stats();
    } elsif ($qry =~ /^split$/) {	# sublist split table
      list_split();
    } elsif ($qry =~ /^([-\w\._]+@?[-\w\._]+)$/) {
      list_info($1);			# a specific list
    } else {			# do status if the args don't make sense
      status();
    }
  }
  exit(0);

############################### Default Status function #####################
sub status() {
  my($nosubs);

  $dbh = db_connect();		# connect to db
  $timenow = get_timenow($dbh);	# get current time at db server

				# Put sendout times into $sent{msg} = time
  make_sent($dbh);
  make_name($dbh,$msgmin,$msgmax);
  $nosubs = get_subno($dbh);

  foreach $n (sort keys %name) {
    $m = $n;			# real listno
    if ($name{$n} =~ /^\d+$/) {
      $m = $name{$n};		# logged listno
    }
    foreach $j (sort {$b <=> $a} keys %sent) {		# go through msgs
      if ($j >= $lo{$n} && $j <= $hi{$n}) {
        $code{$m}[$j]{done} = -2;			# no status
      } elsif (!$code{$m}[$j]{done}) {
        $code{$m}[$j]{done} = -3;			# not active
      }
    }
  }

########## data collection and processing (1/2)
  $sth = $dbh->prepare(	"SELECT msgnum,listno,".
			"max(done),unix_timestamp(max(tai)) AS maxdone ".
			"FROM ${troot}_mlog ".
			"WHERE msgnum BETWEEN $msgmin AND $msgmax ".
			"GROUP BY msgnum,listno ".
			"ORDER BY listno,msgnum DESC")
	or print_error($dbh->errstr);
  $sth->execute or print_error("Getting status info: " . $dbh->errstr);
  $rv = $sth->bind_columns(undef, \$msg, \$ln, \$done, \$tai)
	or print_error("Binding status info: " . $dbh->errstr);
  while($sth->fetch) {
    if ($name{$ln} =~/^\d+$/) {			# pointer to lower list
      $ln = $name{$ln};
    }
    if ($code{$ln}[$msg]{done} == -2) {		# need data for these
      $code{$ln}[$msg]{done} = $done;
      $code{$ln}[$msg]{tai} = $tai;
    }
  }
  $sth->finish;						# ignore errors
  $dbh->disconnect;

########## html start
  content_type("html");
  print_head("Status of list cluster $pi",$REFRESH);
  print_body();

########## table header
  print "<center><TABLE $TBGCOLOR BORDER=3>\n";
  print "<TR $HBGCOLOR><TH COLSPAN=$nomsg>",
	"<a href=\"$URL/$pi?stats\">Messages $msgmax to $msgmin</a>",
	"<TH ROWSPAN=2><a href=\"$URL/$pi?split\">List name</a>\n",
	"<TR $HBGCOLOR>";
  foreach $i (sort {$b <=> $a} keys %sent) {
    $label = $i % 100;
    if ($i >= 100) {
      $label = "'$label";
    }
    print "<TD><a href=\"$URL/$pi?$i\">$label</a>";
  }

########## table data and processing (2/2)
  foreach $i (keys %name) {
    next if $name{$i} =~ /^\d+$/;	# pointer to lower list
    print "\n<TR $DBGCOLOR>";
    $list = clean_name($name{$i});
    foreach $j (sort {$b <=> $a} keys %sent) {
      $tmp = $code{$i}[$j]{done};
      $dt = $timenow - $sent{$j};
      $ago = 0;
      if ($tmp > -2) {
        $ago = $timenow - $code{$i}[$j]{tai};
      }
      $tproc = $dt - $ago;	# $code{tai} - $sent{$j};
      print "<TD><a href=\"$URL/$pi?" ,
		"${j}_${dt}_${ago}_${tmp}_$list\"><img src=\"$IMAGES/";
      if ($tmp == -3) {				# -3 not active ->clear
	print "clear.gif\" alt=\"...\">";
      } elsif ($tmp == 2) {			# 2 -> green
	print "green.gif\" alt=\"[+]\">";
      } elsif ($tmp == -1) {			# -1 bounced -> red
	print "red.gif\" alt=\"[!]\">";
      } elsif ($tmp == -2) {			# no data
	if ($dt > 600) {			# > 10 min -> red
	  print "red.gif\" alt=\"[!]\">";
	} elsif ($dt > 300) {			# > 5 min -> orange
	  print "orange.gif\" alt=\"[?]\">";
	} else {				# <=5 min -> white
	  print "white.gif\" alt=\"[ ]\">";
	}
      } elsif ($tmp == 0) {			# arrived, not done
	if ($ago > 300 || $dt > 600) {	# >5 min processing or > 10 min sent
	  print "red.gif\" alt=\"[!]\">";
	} elsif ($ago < 60 && $dt < 300) {	# < 1 min proc && < 5 min sent
	  print "yellow.gif\" alt=\"[.]\">";
	} else {
	  print "orange.gif\" alt=\"[?]\">";
	}
     } else {					# finished by list
	if ($ago > 1200) {			# > 20 min since done
	  print "red.gif\" alt=\"[!]\">";
	} elsif ($ago > 600) {			# > 10 min since done
	  print "orange.gif\" alt=\"[?]\">";
	} else {
	  print "yellow.gif\" alt=\"[.]\">";
	}
      }
      print "</a>";
    }
    print "<TD><a href=\"$URL/$pi?$list\">$list</a>";
  }
  print "\n</TABLE></center><HR>\n";

  $datestring = gmtime($sent{$msgmax});
  $timestring = make_timestring($timenow - $sent{$msgmax});
  print "Status for list cluster $pi.<P>";
  print "The latest message was sent on $datestring UTC; $timestring ago.<P>\n";
  $timestring = make_timestring($sent{$msgmax} - $sent{$msgmin});
  if ($sent{$msgmax} != $sent{$msgmin}) {
    print "Data represent traffic for $timestring.<BR>\n";
    $dt = $nomsg*360000 / ($sent{$msgmax} - $sent{$msgmin}); # 100x
    $n = int($nosubs * $dt + 0.5)/100;
    $dt = int($dt + .5)/100;
    print "Average ($nosubs subscribers): $dt messages/h; $n deliveries/h.\n";
  }
  print "</BODY></HTML>\n";
}

###################### message info function ##########################
sub msg_info($) {
  local($msg) = @_;

# Do a summary for this message. Times, etc, for all lists.
# SELECT listno,done,tai
# FROM list_mlog 
# WHERE msgnum = arg
#
# time to start | processing time | time to receipt || listname
#               |                 |                 |

  use integer;
  my($timenow,$dt,$datestring,$timestring);

  $dbh = db_connect();			# connect to db;

  $timenow = get_timenow($dbh);	# get current time at db server
  make_sent($dbh,$msg);			# get message sendout times
  if (!defined($sent{$msg})) {
    print_error("Message $msg has not been sent.");
  }
  $datestring = gmtime($sent{$msg});
  $timestring = make_timestring($timenow - $sent{$msg});
  make_name($dbh,$msg,$msg);
  $sth = $dbh->prepare("SELECT listno,done,unix_timestamp(tai) ".
		"FROM ${troot}_mlog ".
		"WHERE msgnum = $msg ORDER BY listno,done")
	or print_error($dbh->errstr);
  $sth->execute
	or print_error("Getting info on message: " . $dbh->errstr);
  $sth->bind_columns(undef, \$n,\$d,\$t)
	or print_error("Binding info on message: " . $dbh->errstr);
  content_type("html");
  print_head("Info for message $msg on $pi");
  print_body();
  print "<CENTER><TABLE $TBGCOLOR BORDER=3>\n",
	"<TR $HBGCOLOR><TH COLSPAN=5>",
	"Message $msg sent: $datestring (UTC); $timestring ago\n";
  print "<TR $HBGCOLOR><TH COLSPAN=4>Seconds elapsed to entry for:",
	"<TH ROWSPAN=2><a href=\"$URL/$pi?split\">List name</a>\n";
  print "<TR $HBGCOLOR>",
	"<TD>Bounce<TD>Arrival<TD>Process<TD>Distribute\n";
  $rc = $sth->fetch;
  while ($rc) {
    print "\n<TR $DBGCOLOR>";
    $oldn = $n;
    $list = clean_name($name{$n});
    for ($i = -1; $i <= 2; $i++) {
      if ($rc) {
	if ($d > 2 || $d < -1) { print_error("Illegal mlog status code."); }
        if ($oldn == $n && $d == $i) {
	  $dt = $t - $sent{$msg};
	  print "<TD ALIGN=RIGHT>$dt";
	  $rc = $sth->fetch;
	} else {
	  print "<TD ALIGN=RIGHT>N/A";
   	}
      } else {
	  print "<TD ALIGN=RIGHT>N/A";
      }
    }
    print "<TD><a href=\"$URL/$pi?$list\">$list</a>";
  }
  print "\n</TABLE></center><HR>\n",
	"Info for message $msg on list cluster ",
	"<a href=\"$URL/$pi\">$pi</a></H1>\n",
	"</BODY></HTML>\n";
}

######################## stats function ############################

sub stats() {

  $dbh = db_connect();
  $timenow = get_timenow($dbh);	# get current time at db server

  make_sent($dbh);		# Put sendout times into $sent{msg} = time
  make_name($dbh,$msgmin,$msgmax);

  foreach $n (0, keys %name) {
    next if ($name{$n} =~ /^\d+$/);			# points to lower list
			# status code here is table + 1 to avoid neg indices!
    for ($d = 0; $d <= 3; $d++) {			# status codes
      foreach $msg (sort {$b <=> $a} keys %sent) {	# go through msgs
	$logged{$n}[$d]{tai}[$msg] = -1;		# no info
      }
    }
  }

  $sth = $dbh->prepare(	"SELECT listno,done,msgnum,unix_timestamp(tai) ".
			"FROM ${troot}_mlog ".
			"WHERE msgnum BETWEEN $msgmin AND $msgmax ".
			"ORDER BY listno,done,msgnum");
  $sth->execute or print_error("Getting mlog: " . $dbh->errstr);
  $sth->bind_columns(undef, \$n,\$d,\$msg,\$tai)
	or print_error("Binding mlog: " . $dbh->errstr);
  while ($sth->fetch) {
			# status code here is table + 1 to avoid neg indices!
    if ($name{$n} =~ /^\d+$/) {
      $n = $name{$n};					# pointer to lower list
    }
    $logged{$n}[$d + 1]{tai}[$msg] = $tai;		# collect timestamps
  }
  $sth->finish;						# ignore errors
  $dbh->disconnect;

########## Document
  content_type("html");
  print_head("Stats on $pi for messages $msgmin to $msgmax");
  print_body();

  print "<center><TABLE $TBGCOLOR BORDER=3>\n",
	"<TR $HBGCOLOR><TH COLSPAN=9>",
	"Summary for messages $msgmin to $msgmax",
	" on list cluster <a href=\"$URL/$pi\">$pi</a>";
  print "<TR $HBGCOLOR><TH COLSPAN=2>to Bounce (s)<TH COLSPAN=2>to Arrival (s)",
	"<TH COLSPAN=2>for Processing (s)<TH COLSPAN=2>to Receipt (s)",
	"<TH ROWSPAN=2><a href=\"$URL/$pi?split\">Listname</a>",
	"<TR $HBGCOLOR>\n";
  for ($i = 1;$i<=4;$i++) { print "<TH>avg<TH>range"; }


  foreach $n (keys %name) {
    next if $name{$n} =~ /^\d+$/;			# pointer to lower list
    print "\n<TR $DBGCOLOR>";
    for ($d = 0; $d <= 3; $d++) {			# status codes
      $tmin = -1;					# min time
      $tmax = -1;					# max time
      $tsum = 0;					# total time
      $tno = 0;						# items
      foreach $msg (sort {$a <=> $b} keys %sent) {	# go through msgs
	$tmp = $logged{$n}[$d]{tai}[$msg];
	if ($tmp != -1) {				# have data
	  if ($d != 2) {
	    $t = $tmp - $sent{$msg};
	  } else {
	    $t = $tmp - $logged{$n}[$d - 1]{tai}[$msg];
	  }
	  $tsum += $t;
	  $tno++;
	  if ($t < $tmin || $tmin == -1) {
	    $tmin = $t;
	  }
	  if ($t > $tmax) {
	    $tmax = $t;
	  }
	}
      }
      if ($tno > 0) {					# got some data
	$tavg = int(($tsum / $tno) * 10 + 0.5)/10;
	  print "<TD ALIGN=RIGHT>$tavg<TD ALIGN=RIGHT>$tmin - $tmax";
	} else {
	  print "<TD ALIGN=RIGHT>-<TD ALIGN=RIGHT>- ";
	}
      }
      $list = clean_name($name{$n});
      print "<TD><a href=\"$URL/$pi?$list\">$list</a>";
    }
    print "\n</TABLE></center><HR>\n",
	"<B>\"Time to arrival\"</B>\n",
	"is the time from message arrival at the main list to start of\n",
	"processing at the sublist. <B>\"Time for processing\"</B> is\n",
	"the time taken at the sublist from message arrival until the\n",
	"message has been handed of to qmail. <B>\"Time to receipt\"</B>\n",
	"is the time from message arrival at the main list until the\n",
	"arrival of the delivery receipt from the sublist. The delivery\n",
	"receipt is sent after trying delivery to all remote subscribers\n",
	"at least once.<P>\n";
    $timestring = make_timestring($timenow - $sent{$msgmax});
    $datestring = gmtime($sent{$msgmax});
    print "The latest message was sent on $datestring UTC; ",
		"$timestring ago.<P>\n";
    $datestring = gmtime($sent{$msgmin});
    $timestring = make_timestring($timenow - $sent{$msgmin});
    print "The first message in this range was sent on $datestring UTC;",
	" $timestring ago.<BR>\n",
	"</BODY></HTML>\n";
    exit(0);
}

######################## list info function ###########################
# show details for this list
# SELECT list_name.listno,msgnum,done,tai
# FROM  list_mlog JOIN list_name ON (listno)
# WHERE name = '$name' AND
# msgnum BETWEEN msgmin AND msgmax
# ORDER BY list_name.listno,done,msgnum

sub list_info($) {
  my ($list) = @_;
  my ($dbh,$sth,$n);
  my ($oldmsg,$oldn,$dnext,$d,$msg,$tai);
  my ($timenow);
  $dbh = db_connect();

  $timenow = get_timenow($dbh);	# get current time at db server
  make_sent($dbh);		# Put sendout times into $sent{msg} = time

  $sth = $dbh->prepare("SELECT t1.listno,msgnum,done,".
		"unix_timestamp(tai) ".
		"FROM ${troot}_name AS t1,${troot}_mlog AS t2 ".
		"WHERE t1.listno = t2.listno AND ".
		"name = '$list' AND ".
		"msgnum BETWEEN $msgmin AND $msgmax ".
		"ORDER BY t1.listno,msgnum,done")
	or print_error($dbh->errstr);
  $sth->execute or print_error("Getting list_info: " . $dbh->errstr);
  $sth->bind_columns(undef, \$n, \$msg, \$d, \$tai)
	or print_error("Binding list_info: " . $dbh->errstr);

########## Document
  content_type("html");
  print_head("Stats on list $list of cluster $pi");
  print_body();

  print "<CENTER><TABLE $TBGCOLOR BORDER=3>\n";
  print "<TR $HBGCOLOR><TH>msg<BR>no<TH>sent (UTC)<TH>list<BR>no",
	"<TH>to Bounce<BR>(s)<TH>to Arrival<BR>(s)",
	"<TH>to Finished<BR>(s)<TH>to Receipt<BR>(s)\n";
  $oldmsg = -1;
  $oldn = -1;
  $dnext = 3;
		# listno, msg -> bounce, msg -> 0 -> 1 -> 2
  while ($sth->fetch) {
    if ($msg != $oldmsg || $n != $oldn) {
      while ($dnext++ <= 2) {
	print "<TD ALIGN=RIGHT>n/a";
      }
      $datestring = gmtime($sent{$msg});
      print "\n<TR $DBGCOLOR><TD ALIGN=RIGHT>",
		"<a href=\"$URL/$pi?$msg\">$msg</a>",
		"<TD>$datestring<TD ALIGN=RIGHT>$n";
      $dnext = -1;
      $oldmsg = $msg;
      $oldn = $n;
    }
    while ($dnext++ < $d) {
      print "<TD ALIGN=RIGHT>n/a";
    }
    print "<TD ALIGN=RIGHT>", ($tai - $sent{$msg});
    $dnext = $d + 1;
  }
  while ($dnext <= 2) {
    print "<TD ALIGN=RIGHT>n/a";
    $dnext++;
  }
  print "\n</TABLE></center>\n<HR>";
  print "Stats on list $ list of cluster ",
	"<a href=\"$URL/$pi\">$pi</a></H1>\n",
	"</BODY></HTML>\n";
  $sth->finish;
  $dbh->disconnect;
}

######################### list_split function #########################
# show the sublist split table of currently active lists (last msg)
sub list_split() {
  my ($dbh,$n,$namr,$dl,$dh,$hl,$hh,$drange,$hrange,$hadd);

  $dbh = db_connect();
  make_sent($dbh);
  $sth = $dbh->prepare("SELECT listno,name,domain,hash_lo,hash_hi ".
		"FROM ${troot}_name ".
		"WHERE $msgmax BETWEEN msgnum_lo AND msgnum_hi ".
		"AND notuse = 0 ".
		"ORDER BY domain,hash_lo,hash_hi,name")
	or print_error($dbh->errstr);
  $sth->execute or print_error("Getting split: " . $dbh->errstr);
  $sth->bind_columns(undef, \$n, \$name, \$d, \$hl, \$hh)
	or print_error("Binding split: " . $dbh->errstr);
  content_type("html");
  print_head("Sublist split table for $pi");
  print_body();
  print	"<center><TABLE $TBGCOLOR BORDER=3>",
	"<TR $HBGCOLOR><TH>Domain<TH>Hash range<TH>List Name";
  while ($sth->fetch) {
    $hrange = "";
    $hadd = "";
    if ("$d" eq "") {
      $d = '-other-';
      $dother = 1;
    }
    if ($hl <= 0) {
      $hl = 0;
    }
    if ($hl > $hh) {
      $hrange = "NONE!";
    } elsif ($hh >= 52) {
      if ($hh == 99) {
	$hadd = "[sublists]";
	$hh = 52;
      }
      if ($hl == 0) {
	$hrange = "-- any --";
	$hany = 1;
      } else {
	if ($hl == $hh) {
	  $hrange = "$hl";
	} elsif ($hl != 99) {
	  $hrange = "$hl and above";
	}
      }
    } else {
      if ($hl == 0) {
	$hrange = "up to $hh";
      } else {
	$hrange = "$hl to $hh";
      }
    }
    print "\n<TR $DBGCOLOR><TD ALIGN=CENTER>$d<TD ALIGN=CENTER>$hrange";
    if ($hadd ne "") {
      if ($hrange ne "") {
	print " &amp; ";
      }
      print $hadd;
    }
    $list = clean_name($name);
    print "<TD><a href=\"$URL/$pi?$list\">$list</a>";
  }
  $sth->finish;						# ignore errors
  $dbh->disconnect;
  print "\n</TABLE></center><HR>\n";
  print "This is the current (message $msgmax) sublist split for",
	" <a href=\"$URL/$pi\">$pi</a><P>\n";
  if ($dother == 1) {
    print "<P>other: Domains not serviced by other sublists.<BR>\n";
  }
  if ($hany == 1) {
    print "any: Entire normal hash range 0-52\n";
  }
  print "</BODY></HTML>\n";
}

######################## explain function #############################
sub expl($$$$$) {
  local($msg,$dt,$ago,$code,$list) = @_;

  content_type("html");

  print "<html><head>\n";
  print "<title>Info for message $msg on sublist $list</title>\n";
  print "</HEAD>\n";
  print_body();
  print "<center><TABLE $TBGCOLOR BORDER=3>\n",
	"<TR $HBGCOLOR><TH ALIGN=CENTER COLSPAN=2>",
	"Message $msg on sublist $list\n";
  print "<TR><TH ALIGN=LEFT $HBGCOLOR>Last logged action:",
	"<TD ALIGN=RIGHT $DBGCOLOR>";
  if ($code <= -2) {
    print "none";
  } elsif ($code == -1) {
    print "bounce";
  } elsif ($code == 0) {
    print "start processing";
  } elsif ($code == 1) {
    print "finished processing";
  } else {
    print "receipt received";
  }
  if ($code > -2) {
    $tproc = $dt - $ago;
    print "\n<TR><TD ALIGN=LEFT $HBGCOLOR>Time taken (s):",
	"<TD ALIGN=RIGHT $DBGCOLOR>$tproc\n";
  } 
  print "\n<TR><TD ALIGN=LEFT $HBGCOLOR>Time since sent (s):",
	"<TD ALIGN=RIGHT $DBGCOLOR>$dt\n";
  print "</TABLE></center>\n",
	"<P><HR>\n";
  
  if ($code == -3) {
    print "This sublist has not received the message, and should not since\n",
	"it is not active (for this message).\n<P>",
	"If this is not what you expect, correct the\n",
	"sublist message split in table ${troot}_name.\n";
  } elsif ($code == -2) {
    print "The sublist has not received the message.<P>\n";
    if ($dt < 300) {
      print "This is normal for a few minutes after the message has\n",
		"arrived at the main list.<P>\n";
    } else {
      print "A delay of more than a few minutes indicates a failure\n",
	"of the main list, or mail transport to the sublist. If the\n",
	"main list receipt has been received, it indicates failure of\n",
	"the mail path between main list and sublist or of qmail at the\n",
	"sublist.<P>\n",
	"An inability to contact the SQL server has the same result.\n",
	"<P>A failure to subscribe the sublist to the main list has the\n",
	"same result.\n",
	"<P>Investigate the qmail log at the sublist!\n";
    }
  } elsif ($code == -1) {
    print "The message to the sublist has bounced:<P>\n",
	"Qmail at the main list was unable to deliver the message to the\n",
	"sublist.<P>There is a permanent problem in the mail path, or\n",
	"the list configuration.\n",
	"<P>Investigate the qmail log at the main list and sublist!\n",
	"Consider disabling this list temporarily to route messages via\n",
	"other sublists!";
  } elsif ($code == 0) {
    print "The sublist has received the message and started to process it.\n";
    if ($ago > 60) {	# allow 60 secs
      print "Processing usually takes no more than a few seconds. Likely,\n",
	"there is a failure at the level of the\n",
	"sublist itself or outgoing qmail at the sublist host.\n",
	"<P>Investigate the qmail log at the sublist!\n";
    }
  } elsif ($code == 1) {
    print "The sublist has successfully processed the message.<P>\n",
	"The distribution receipt has not yet been received.<P>\n",
	"Delivery may not yet have been attempted at least once to each\n",
	"subscriber.<P>\n";
    if ($ago > 1200) {
      print "This is taking quite long. Likely,\n",
	"there is an outgoing mail failure at the sublist, either at the\n",
	"level of qmail or in the mail path for the feedback message back\n",
	"to the main list.\n",
	"<P>Investigate the qmail log at the sublist!\n";
    }
  } else {
    print "Everything is in order:<P>\n",
	"The message has been successfully processed at the sublist.<P>\n",
	"Delivery has been attempted at least once to each subscriber.<P>\n",
	"The sublist host is able to successfully send out mail.\n";
  }
  print "<P>\n";
  print "</body></html>\n";
}

########################## DATABASE SUBS #########################

########### Connect to database server ###
sub db_connect() {
  $dbh_ = DBI->connect("DBI:$DRIVER:$db:$host",$user,$pw)
    or print_error("Failed to connect to (or authenticate with) SQL server." .
	"Most likely, the DSOURCE setting in the script is not correct.");
  return $dbh_;
}

########### Get current time at the database server ###
sub get_timenow($) {
  my ($dbh) = @_;
  my $tn,$sth;

  $sth = $dbh->prepare("SELECT unix_timestamp(current_timestamp())")
	or print_error($dbh->errstr);
  $sth->execute
	or print_error("Getting current time: " . $dbh->errstr);
  (($tn) = $sth->fetchrow_array)
	or print_error("Fetching current time: " . $dbh->errstr);
  $sth->finish
	or print_error($dbh->errstr);
  return $tn;
}

###################### setup %sent msg->tai ###################################
sub make_sent($;$) {
  my ($dbh,$msg) = @_;
  my $i,$sth,$m,$t;

  $msgmin = 0;
  $msgmax = 0;
  $nomsg = 0;
  if ($msg) {
    $sth = $dbh->prepare("SELECT msgnum,unix_timestamp(tai) FROM ".
			"${troot}_cookie ".
			"WHERE msgnum = $msg")
	or print_error($dbh->errstr);
  } else {
    $sth = $dbh->prepare("SELECT msgnum,unix_timestamp(tai) FROM ".
			"${troot}_cookie ".
			"ORDER by msgnum DESC limit $NOMSG")
	or print_error($dbh->errstr);
  }
  $sth->execute
	or print_error("Getting message sendout times: " . $dbh->errstr);
  $rv = $sth->bind_columns(undef,\$m, \$t)
	or print_error("Binding message sendout times: " . $dbh->errstr);
  while ($sth->fetch) {
    $nomsg++;
    $sent{$m} = $t;
    $msgmax = $m if $msgmax == 0;	# first is max
    $msgmin = $m;			# last is min
  }
  $sth->finish
	or print_error($dbh->errstr);
}

###################### get subscriber no  #####################################
sub get_subno($) {
  my ($dbh) = @_;
  my $tn;
  $sth = $dbh->prepare("SELECT count(*) from ".
			"${troot} ")
	or print_error($dbh->errstr);
  $sth->execute
	or print_error("Getting subscriber no: " . $dbh->errstr);
  (($n) = $sth->fetchrow_array)
	or print_error("Fetching subscriber no: " . $dbh->errstr);
  $sth->finish
	or print_error($dbh->errstr);
  return $n;
}

####################### setup %name/lo/hi #####################################
sub make_name($$$) {
  my ($dbh,$msgmin,$msgmax) = @_;
  local ($n,$i,$j,$hi,$lo,$sth);
  local ($prevname) = ("");

  $sth = $dbh->prepare("SELECT listno,name,msgnum_lo,msgnum_hi ".
			"FROM ${troot}_name WHERE notuse = 0 AND ".
			"$msgmax >= msgnum_lo AND $msgmin <= msgnum_hi ".
			"ORDER BY name,listno,msgnum_lo,msgnum_hi ")
	or print_error($dbh->errstr);
  $sth->execute or print_error("Getting lists: " . $dbh->errstr);
  $sth->bind_columns(undef, \$n,\$listname,\$lo,\$hi)
	or print_error("Binding lists: " . $dbh->errstr);
  FETCHLOOP: while ($sth->fetch) {		# info for one list
    if ($lo < $msgmin) { $lo = $msgmin; }	# restrict to range
    if ($hi > $msgmax) { $hi = $msgmax; }
    if ("$listname" eq "$prevname") {		# same name
      foreach $tl (@thislist) {			# make non-overlapping
	next FETCHLOOP if ($lo == $lo{$tl} && $hi == $hi{$tl});
      }
      push @thislist,$n;
      $name{$n} = $thislist[0];
    } else {
      $prevname = $listname;
      $name{$n} = $listname;			# name hash
      undef @thislist;
      push @thislist,$n;
    }
    $lo{$n} = $lo;
    $hi{$n} = $hi;
  }
  $sth->finish or print_error($dbh->errstr);
}

########################## MISC SUBS #########################
########### print error status message
sub print_error($) {
  local($errstr) = @_;
  content_type("html","500 Script error");
  print "<HTML><HEAD><TITLE>Status: 500 Script Error</TITLE></HEAD>\n",
	"<BODY><H1>Script Error</H1>\n",
	"<H2>Execution of the script failed with the following error:</H2>\n";
  print "$errstr\n<P></BODY></HTML>";
  exit(0);
}

########### make time string

sub make_timestring($) {
  local ($s) = @_;
  my($t);
  if ($s < 120) {
    return "$s s";
  } elsif ($s < 7200) {
    $t = int($s/6+0.5)/10;
    return "$t min";
  } elsif ($s < 172800) {
    $t = int($s/360+0.5)/10;
    return "$t h";
  } else {
    $t = int($s/864+0.5)/100;
    return "$t days";
  }
}

########### remove offensive characters from list name
sub clean_name($) {
  local($n) = @_;
  $n =~ tr/-+0-9a-zA-Z@\./_/cs;
  return $n;
}

########### print content-type header ###
sub content_type($;$) {
  local($s,$status) = @_;
  print "Content-Type: text/$s\n";
  if (!$status) {
    print "\n";
  } else {
    print "Status: $status\n\n";
  }
}

sub print_head($;$) {
  my ($title,$refresh) = @_;
  print "<HTML><HEAD>\n";
  print "<META HTTP-EQUIV=\"PRAGMA\" CONTENT=\"NO-CACHE\">\n";
  if ($refresh) {
    print "<META HTTP-EQUIV=\"REFRESH\" CONTENT=\"$refresh\">\n";
  }
  print "<TITLE>$title</TITLE>\n";
  print "</HEAD>\n";
}

########## print doc body tag ###
sub print_body() {
  print "<BODY $BBGCOLOR $TEXT $LINK $VLINK>\n";
}

